structure Expr:> EXPR = struct
  datatype t =
    Var of string
  | Int of int
  | Str of string
  | Plus of t * t
  | Concat of t * t
  | Len of t
  | ToStr of t
  | Let of string * t * t

  fun toString (e: t): string =
    case e of
      Var x => x
    | Int n => Int.toString n
    | Str s => "\"" ^ s ^ "\""
    | Plus (e1, e2) => "(" ^ toString e1 ^ " + " ^ toString e2 ^ ")"
    | Concat (e1, e2) => "(" ^ toString e1 ^ " ^ " ^ toString e2 ^ ")"
    | Len e1 => "len(" ^ toString e1 ^ ")"
    | ToStr e1 => "toString(" ^ toString e1 ^ ")"
    | Let (x, e1, e2) => "let " ^ x ^ " = " ^ toString e1 ^ " in " ^ toString e2

  fun subst (e1: t) (x: string) (e2: t): t =
    case e2 of
      Var y => if x = y then e1 else Var y
    | Int n => Int n
    | Str s => Str s
    | Plus (e3, e4) => Plus (subst e1 x e3, subst e1 x e4)
    | Concat (e3, e4) => Concat (subst e1 x e3, subst e1 x e4)
    | Len e3 => Len (subst e1 x e3)
    | ToStr e3 => ToStr (subst e1 x e3)
    | Let (y, e3, e4) =>
        (* if e1 has no free variables, then this will not incur capture. *)
        Let (y, subst e1 x e3, if x = y then e4 else subst e1 x e4)
end
